#!/usr/bin/python
# -*- coding: utf-8 -*-
 #
 # stinkpad(a)blueyonder.co.uk	2007-11-26
 """Toggle internal and external displays (equivalent to ThinkPad Fn7)
 Simple; cloned: on+off, on+on, off+on. 
 Full; as simple plus xinerama: right, below,  left, above. 
 Limited; as full but display will overlap if virtual screen is too small. 
 Use 'xrandr -q' to determine output names. 
 Further information: http://www.thinkwiki.org/wiki/Xorg_RandR_1.2 
 """
 __usage__ = "usage: %prog [--help]|[[-i internal][-e external][-d displays]]"
 __version__ = "toggle [djclark.eu 2007-11-26]"
 #
 # Output names; Intel: LVDS VGA TV TMDS-1 TMDS-2
 #               ATI:   LVDS VGA-0 S-video DVI-0
 LAPTOP = 'LVDS' 
 MONITOR = 'VGA'
 SEQUENCE = 'simple'
 #
 import sys
 import os
 import re
 
 # "LVDS connected 1024x768+0+0 (normal left inverted right) 304mm x 228mm"
 REGEX_OUTPUT = re.compile(r"""
 	(?x)					# ignore whitespace
 	^					# start of string
 	(?P<output>[A-Za-z0-9\-]*)[ ] 		# LVDS VGA etc
 	(?P<connect>(dis)?connected)[ ]		# dis/connected
 	((					# a group
  	(?P<width>\d+)x 			# either 1024x768+0+0
 	(?P<height>\d+)[+]  
  	(?P<horizontal>\d+)[+]
 	(?P<vertical>\d+)
 	)|[\D])					# or not a digit
 	.*					# ignore rest of line
 	""")
 
 # "Screen 0: minimum 320 x 200, current 1024 x 768, maximum 2624 x 1968"
 REGEX_SCREEN = re.compile(r"""
 	(?x) 				# ignore whitespace
 	^				# start of string
 	Screen[ ]			
 	(?P<screen>\d)[: ]+
 	minimum[ ]
 	(?P<minWidth>\d+)[ x]+
 	(?P<minHeight>\d+)[, ]+
 	current[ ]+
 	(?P<curWidth>\d+)[ x]+
 	(?P<curHeight>\d+)[, ]+
 	maximum[ ]+
 	(?P<maxWidth>\d+)[ x]+
 	(?P<maxHeight>\d+)
 	""")	
 
 def toggle_simple(d0, d1):
    """Toggle display states: on+off, on+on, off+on"""
    if d1['connect'] == 'disconnected': 	# external unplugged
        return ('auto','off',"") 		#     switch off external
    if d1['width'] is 0: 			# external off
        return ('auto','auto',"") 		#     both on
    if d0['width'] is 0: 			# laptop off
        return ('auto','off',"") 		#    laptop on
    return ('off','auto',"") 			# both on, laptop off
 
 def toggle_full(d0, d1):
    """Toggle display states: 1+0, 1+1, 0+1, 1+E, 1+S, 1+W, 1+N""" 
    if d1['connect'] == 'disconnected': 	# external unplugged
        return ('auto','off',"") 		#     switch off external
    place = '--%s ' + d0['output']
    if d1['width'] == 0: 			# external off
        return ('auto','auto',place%'same-as') 	#     external on
    if d0['width'] == 0: 			# laptop off
        return ('auto','off',"") 		#    laptop on
    if d1['horizontal'] > 0: 			# external to right
        return ('auto','auto',place%'below') 	#     make below
    if d1['vertical'] > 0: 			# external below
        return ('auto','auto',place%'left-of') 	#     make left
    if d0['horizontal'] > 0: 			# external left
        return ('auto','auto',place%'above') 	#     make above
    if d0['vertical'] > 0: 			# external above
        return ('off','auto',"")  		#     laptop off
    return ('auto','auto',place%'right-of') 	# is same, make right
 
 def position(orientation, da, db, screen):
    """Calculate offset position of second display"""
    p = 'auto --pos %sx%s'
    if orientation == 'V':
        if da['height'] + db['height'] <= screen['maxHeight']:
            return p%(0, da['height'])
        return p%(0, screen['maxHeight'] - db['height'])
    else:
        if da['width'] + db['width'] <= screen['maxWidth']:
            return p%(da['width'],0)
        return p%(screen['maxWidth'] - db['width'],0)
 
 def toggle_limited(d0, d1, sz):
    """Toggle display states (overlapped): 1+0,1+1,0+1,1+E,1+S,1+W,1+N"""
    if d1['connect'] == 'disconnected': 	# external unplugged
        return ('auto','off') 			#     switch off external
    if d1['width'] == 0: 				# external off
        return ('auto --pos 0x0','auto --pos 0x0') 	#     both on
    if d0['width'] == 0: 				# laptop off
        return ('auto --pos 0x0','off') 		#     laptop on
    if d1['horizontal'] > 0: 				# external to right
        return ('auto --pos 0x0',position('V',d0,d1,sz)) #     put *below
    if d1['vertical'] > 0: 				# external below
        return (position('H',d1,d0,sz),'auto --pos 0x0') #     put *left
    if d0['horizontal'] > 0: 				# external left
        return (position('V',d1,d0,sz),'auto --pos 0x0') #     put *above
    if d0['vertical'] > 0: 				# external above
        return ('off','auto --pos 0x0')  		#     laptop off
    return ('auto --pos 0x0',position('H',d0,d1,sz)) 	# both, put*right
 
 class DisplayNameError(UnboundLocalError):
    """Internal or External Display Name not found by xrandr -q """
 
 def function7(disp0=LAPTOP, disp1=MONITOR, seq=SEQUENCE):
    """Use xrandr to read current display state and change state"""
    for line in os.popen('xrandr -q'):
        if line.startswith(disp0,0) :
            d0_state = REGEX_OUTPUT.match(line).groupdict()
        elif line.startswith(disp1,0):
            d1_state = REGEX_OUTPUT.match(line).groupdict()
        elif line.startswith('Screen',0):
            screen_size = REGEX_SCREEN.match(line).groupdict()
    for i in ('width','height','horizontal','vertical'):
        try:
            d0_state[i] = int(d0_state[i])
        except TypeError:
            d0_state[i] = 0
        except UnboundLocalError:
            raise DisplayNameError, 'Internal Display: %s not found'% disp0
        try:
            d1_state[i] = int(d1_state[i])
        except TypeError:
            d1_state[i] = 0
        except UnboundLocalError:
            raise DisplayNameError, 'External Display: %s not found'% disp1
    for i in screen_size:
        try:
            screen_size[i] = int(screen_size[i])
        except TypeError:
            screen_size[i] = 0
    #
    toggle = toggle_simple
    xrandr ='xrandr --output '+disp0+' --%s --output '+disp1+' --%s %s'
    if seq == 'full':
        toggle = toggle_full
    if seq == 'limited':
        toggle = toggle_limited
        xrandr ='xrandr --output '+disp0+' --%s --output '+disp1+' --%s'
        os.popen(xrandr % toggle(d0_state, d1_state, screen_size))
    else:
        os.popen(xrandr % toggle(d0_state, d1_state))
 
 def main():
    """ Command line options """
    from optparse import OptionParser
    p = OptionParser(usage=__usage__, version=__version__, description=__doc__)  
    p.set_defaults(internal=LAPTOP, external=MONITOR, displays=SEQUENCE)
    p.add_option('-i','--internal', dest="internal", metavar=LAPTOP,
 	help="internal display")
    p.add_option('-e','--external', dest="external", metavar=MONITOR,
  	help="external display")
    p.add_option('-d','--displays', dest="displays", action="store",
 	choices=('simple', 'limited', 'full'), metavar=SEQUENCE,
 	help='simple/limited/full')
    (opt, args) = p.parse_args()
    try:
        function7(opt.internal, opt.external, opt.displays)
    except DisplayNameError, err:
        print '\n'+str(err)+'\n'
        print os.popen('xrandr -q').read()
 #
 if __name__ == '__main__': 	#only when run from cmd line
    main()
